/*
 * The OpenSAML License, Version 1.
 * Copyright (c) 2002
 * University Corporation for Advanced Internet Development, Inc.
 * All rights reserved
 *
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution, if any, must include
 * the following acknowledgment: "This product includes software developed by
 * the University Corporation for Advanced Internet Development
 * <http://www.ucaid.edu>Internet2 Project. Alternately, this acknowledegement
 * may appear in the software itself, if and wherever such third-party
 * acknowledgments normally appear.
 *
 * Neither the name of OpenSAML nor the names of its contributors, nor
 * Internet2, nor the University Corporation for Advanced Internet Development,
 * Inc., nor UCAID may be used to endorse or promote products derived from this
 * software without specific prior written permission. For written permission,
 * please contact opensaml@opensaml.org
 *
 * Products derived from this software may not be called OpenSAML, Internet2,
 * UCAID, or the University Corporation for Advanced Internet Development, nor
 * may OpenSAML appear in their name, without prior written permission of the
 * University Corporation for Advanced Internet Development.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND WITH ALL FAULTS. ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, AND NON-INFRINGEMENT ARE DISCLAIMED AND THE ENTIRE RISK
 * OF SATISFACTORY QUALITY, PERFORMANCE, ACCURACY, AND EFFORT IS WITH LICENSEE.
 * IN NO EVENT SHALL THE COPYRIGHT OWNER, CONTRIBUTORS OR THE UNIVERSITY
 * CORPORATION FOR ADVANCED INTERNET DEVELOPMENT, INC. BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */


/* saml.h - OpenSAML header file

   Scott Cantor
   5/7/02

   $History:$
*/

#ifndef __saml_h__
#define __saml_h__

#include <xercesc/dom/DOM.hpp>
#include <xercesc/framework/Wrapper4InputSource.hpp>
#include <xercesc/sax/InputSource.hpp>
#include <xercesc/util/BinInputStream.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/util/XMLDateTime.hpp>
#include <xsec/dsig/DSIGSignature.hpp>

using namespace xercesc;

#ifdef WIN32
# include <windows.h>
# ifndef SAML_EXPORTS
#  define SAML_EXPORTS __declspec(dllimport)
# endif
# pragma warning(disable : 4786 4231 4250 4251 4503 4661)
#else
# define __stdcall
# define __cdecl
# define SAML_EXPORTS
#endif

/* Xerces defines its own macro for wide literals, but on platforms without 16 bit wchar_t
   it defines away the L"" and assumes the ASCII literal is usable, which is never true.
   I convert L"foo" to L(foo) so that the macro handles the reference to the static constant,
   while the code stays somewhat similar.
 */
#define L(s) saml::XML::Literals::s
#define L_QNAME(p,s) saml::XML::Literals::p##_##s

/* Implement a class name mechanism for use with logging. */
# define RTTI(type) _RTTI(#type)

// STL headers
#include <map>
#include <set>
#include <stack>
#include <string>
#include <vector>
#include <iostream>
#include <exception>

#if defined (_MSC_VER) || defined(__BORLANDC__)
# include <saml/config_pub_win32.h>
#else
# include <saml/config_pub.h>
#endif

namespace saml
{
#ifdef HAVE_GOOD_STL
    // A dynamic widestring that works across platforms
    typedef std::basic_string<XMLCh> xstring;
#endif

    // Wraps a UTF-8 transcoder (both directions).
    SAML_EXPORTS char* toUTF8(const XMLCh* src);
    SAML_EXPORTS XMLCh* fromUTF8(const char* src);

    // Can't use std::auto_ptr on some platforms so we need similar classes.
    // These are NOT full auto_ptrs, they're just useful for simple cases.
    class SAML_EXPORTS auto_ptr_char
    {
    public:
        auto_ptr_char(const XMLCh* src) : m_buf(XMLString::transcode(src)) {}
        auto_ptr_char(const char* src) : m_buf(XMLString::replicate(src)) {}
        ~auto_ptr_char() { XMLString::release(&m_buf); }
        const char* get() const { return m_buf; }
        char* release() { char* temp=m_buf; m_buf=NULL; return temp; }

    private:
        char* m_buf;
    };

    class SAML_EXPORTS auto_ptr_XMLCh
    {
    public:
        auto_ptr_XMLCh(const char* src) : m_buf(XMLString::transcode(src)) {}
        auto_ptr_XMLCh(const XMLCh* src) : m_buf(XMLString::replicate(src)) {}
        ~auto_ptr_XMLCh() { XMLString::release(&m_buf); }
        const XMLCh* get() const { return m_buf; }
        XMLCh* release() { XMLCh* temp=m_buf; m_buf=NULL; return temp; }

    private:
        XMLCh* m_buf;
    };

    // Java-style Iterator interfaces
    template<class T> class Iterator
    {
    public:
        Iterator();
        Iterator(const std::vector<T>& v) : m_vector(v) { m_iter=m_vector.begin(); }
        std::vector<T> clone() const;
        bool hasNext() const { return m_iter!=m_vector.end(); }
        const T& next() const { return *(m_iter++); }
        size_t size() const { return m_vector.size(); }
        const T& at(size_t i) const { return m_vector[i]; }
        const T& operator[](size_t i) const { return m_vector[i]; }
        void reset() const { m_iter=m_vector.begin(); }

    protected:
        const std::vector<T>& m_vector;
        mutable typename std::vector<T>::const_iterator m_iter;

    private:
        static std::vector<T> m_empty_vector;
    };

    template<class T> class ArrayIterator : public Iterator<T>
    {
    public:
        ArrayIterator(T v[], size_t s);

    protected:
        std::vector<T> m_real_vector;
    };

    #define EMPTY(T) saml::Iterator<T>()

    template<class T> std::vector<T> Iterator<T>::m_empty_vector;

    template<class T> Iterator<T>::Iterator() : m_vector(m_empty_vector)
    {
        m_iter=m_vector.begin();
    }

    template<class T> ArrayIterator<T>::ArrayIterator(T v[], size_t s) : Iterator<T>(m_real_vector)
    {
        for (size_t i=0; v && i<s; i++)
            m_real_vector.push_back(v[i]);
        this->m_iter=this->m_vector.begin();
    }

#ifdef SAML_INSTANTIATE
    class QName;
    class SAMLAttribute;
    class SAMLAttributeDesignator;
    class SAMLAuthorityBinding;
    class SAMLStatement;
    class SAMLCondition;
    class SAMLAssertion;
    class SAMLAction;
    class SAMLEvidence;

#ifdef HAVE_GOOD_STL
    template class SAML_EXPORTS saml::Iterator<xstring>;
    template class SAML_EXPORTS saml::ArrayIterator<xstring>;
#endif
    template class SAML_EXPORTS saml::Iterator<std::string>;
    template class SAML_EXPORTS saml::ArrayIterator<std::string>;
    template class SAML_EXPORTS saml::Iterator<const XMLCh*>;
    template class SAML_EXPORTS saml::ArrayIterator<const XMLCh*>;
    template class SAML_EXPORTS saml::Iterator<const DOMElement*>;
    template class SAML_EXPORTS saml::ArrayIterator<const DOMElement*>;
    template class SAML_EXPORTS saml::Iterator<saml::QName>;
    template class SAML_EXPORTS saml::ArrayIterator<saml::QName>;
    template class SAML_EXPORTS saml::Iterator<XSECCryptoX509*>;
    template class SAML_EXPORTS saml::ArrayIterator<XSECCryptoX509*>;
    template class SAML_EXPORTS saml::Iterator<SAMLAttribute*>;
    template class SAML_EXPORTS saml::ArrayIterator<SAMLAttribute*>;
    template class SAML_EXPORTS saml::Iterator<SAMLAttributeDesignator*>;
    template class SAML_EXPORTS saml::ArrayIterator<SAMLAttributeDesignator*>;
    template class SAML_EXPORTS saml::Iterator<SAMLAuthorityBinding*>;
    template class SAML_EXPORTS saml::ArrayIterator<SAMLAuthorityBinding*>;
    template class SAML_EXPORTS saml::Iterator<SAMLStatement*>;
    template class SAML_EXPORTS saml::ArrayIterator<SAMLStatement*>;
    template class SAML_EXPORTS saml::Iterator<SAMLCondition*>;
    template class SAML_EXPORTS saml::ArrayIterator<SAMLCondition*>;
    template class SAML_EXPORTS saml::Iterator<SAMLAssertion*>;
    template class SAML_EXPORTS saml::ArrayIterator<SAMLAssertion*>;
    template class SAML_EXPORTS saml::Iterator<SAMLAction*>;
    template class SAML_EXPORTS saml::ArrayIterator<SAMLAction*>;
    template class SAML_EXPORTS saml::Iterator<SAMLEvidence*>;
    template class SAML_EXPORTS saml::ArrayIterator<SAMLEvidence*>;
#endif

    // Abstract handling of diagnostic context for logging
    class SAML_EXPORTS NDC
    {
    public:
        NDC();
        NDC(const char* context);
        NDC(std::string& context);
        ~NDC();
    };

    // Abstracts handling of XML QNames, including namespace prefix traversal
    class SAML_EXPORTS QName
    {
    public:
        QName(const XMLCh* nsURI=NULL, const XMLCh* localName=NULL);
        ~QName();

        static QName* getQNameAttribute(DOMElement* e, const XMLCh* nsURI, const XMLCh* localName);
        static QName* getQNameTextNode(DOMText* t);
        static const XMLCh* getNamespaceForQName(const XMLCh* qname, DOMElement* e);
        static const XMLCh* getNamespaceForPrefix(const XMLCh* prefix, DOMElement* e);

#ifdef HAVE_GOOD_STL
        const XMLCh* getNamespaceURI() const {return m_namespace.c_str();}
        const XMLCh* getLocalName() const {return m_localName.c_str();}

    private:
        xstring m_namespace;
        xstring m_localName;
#else
        QName(const QName& src);
        QName& operator=(const QName& src);

        const XMLCh* getNamespaceURI() const {return m_namespace;}
        const XMLCh* getLocalName() const {return m_localName;}

    private:
        XMLCh* m_namespace;
        XMLCh* m_localName;
#endif
    };

    // Needed to allow QNames in sorted STL containers
    SAML_EXPORTS bool operator<(const saml::QName& op1, const saml::QName& op2);
#ifndef HAVE_GOOD_STL
    SAML_EXPORTS bool operator==(const saml::QName& op1, const saml::QName& op2);
#endif
    
    // Wraps pseudorandom identifier generation for SAML objects
    class SAML_EXPORTS SAMLIdentifier
    {
    public:
        SAMLIdentifier();
        ~SAMLIdentifier();
        operator const XMLCh*() {return m_buffer;}

    private:
        SAMLIdentifier(const SAMLIdentifier&) {}
        SAMLIdentifier& operator=(const SAMLIdentifier&) {return *this;}
        XMLCh* m_buffer;
    };

    // Wraps basic XML handling and wide constants (ugly, but fast)
    class SAML_EXPORTS XML
    {
    public:
        // URI constants
        static const XMLCh OPENSAML_NS[];
        static const XMLCh SAML_NS[];
        static const XMLCh SAML_SCHEMA_ID[];
        static const XMLCh SAML11_SCHEMA_ID[];
        static const XMLCh SAMLP_NS[];
        static const XMLCh SAMLP_SCHEMA_ID[];
        static const XMLCh SAMLP11_SCHEMA_ID[];
        static const XMLCh SOAP11ENV_NS[];
        static const XMLCh SOAP11ENV_SCHEMA_ID[];
        static const XMLCh XMLSIG_NS[];
        static const XMLCh XMLSIG_SCHEMA_ID[];
        static const XMLCh XPATH2_NS[];
        static const XMLCh XPATH2_SCHEMA_ID[];
        static const XMLCh XMLNS_NS[];
        static const XMLCh XML_NS[];
        static const XMLCh XML_SCHEMA_ID[];
        static const XMLCh XSD_NS[];
        static const XMLCh XSI_NS[];
        
        // Helper functions and whitespace skippers
        static bool isElementNamed(const DOMElement* e, const XMLCh* ns, const XMLCh* localName);
        static DOMElement* getFirstChildElement(const DOMNode* n);
        static DOMElement* getLastChildElement(const DOMNode* n);
        static DOMElement* getNextSiblingElement(const DOMNode* n);
        static DOMElement* getPreviousSiblingElement(const DOMNode* n);
        static DOMElement* getFirstChildElement(const DOMNode* n, const XMLCh* ns, const XMLCh* localName);
        static DOMElement* getLastChildElement(const DOMNode* n, const XMLCh* ns, const XMLCh* localName);
        static DOMElement* getNextSiblingElement(const DOMNode* n, const XMLCh* ns, const XMLCh* localName);
        static DOMElement* getPreviousSiblingElement(const DOMNode* n, const XMLCh* ns, const XMLCh* localName);

        struct SAML_EXPORTS Literals
        {
            // SAML vocabulary
            static const XMLCh Action[];
            static const XMLCh Advice[];
            static const XMLCh Assertion[];
            static const XMLCh AssertionArtifact[];
            static const XMLCh AssertionID[];
            static const XMLCh AssertionIDReference[];
            static const XMLCh saml_AssertionIDReference[];
            static const XMLCh Attribute[];
            static const XMLCh AttributeDesignator[];
            static const XMLCh AttributeName[];
            static const XMLCh AttributeNamespace[];
            static const XMLCh AttributeQuery[];
            static const XMLCh AttributeQueryType[];
            static const XMLCh AttributeStatement[];
            static const XMLCh AttributeStatementType[];
            static const XMLCh AttributeValue[];
            static const XMLCh Audience[];
            static const XMLCh AudienceRestrictionCondition[];
            static const XMLCh AudienceRestrictionConditionType[];
            static const XMLCh AuthenticationInstant[];
            static const XMLCh AuthenticationMethod[];
            static const XMLCh AuthenticationQuery[];
            static const XMLCh AuthenticationQueryType[];
            static const XMLCh AuthenticationStatement[];
            static const XMLCh AuthenticationStatementType[];
            static const XMLCh AuthorityBinding[];
            static const XMLCh AuthorityKind[];
            static const XMLCh AuthorizationDecisionQuery[];
            static const XMLCh AuthorizationDecisionQueryType[];
            static const XMLCh AuthorizationDecisionStatement[];
            static const XMLCh AuthorizationDecisionStatementType[];
            static const XMLCh Binding[];
            static const XMLCh Condition[];
            static const XMLCh Conditions[];
            static const XMLCh ConfirmationMethod[];
            static const XMLCh Decision[];
            static const XMLCh DNSAddress[];
            static const XMLCh DoNotCacheCondition[];
            static const XMLCh DoNotCacheConditionType[];
            static const XMLCh Evidence[];
            static const XMLCh Format[];
            static const XMLCh InResponseTo[];
            static const XMLCh IPAddress[];
            static const XMLCh Issuer[];
            static const XMLCh IssueInstant[];
            static const XMLCh Location[];
            static const XMLCh MajorVersion[];
            static const XMLCh MinorVersion[];
            static const XMLCh NameIdentifier[];
            static const XMLCh NameQualifier[];
            static const XMLCh Namespace[];
            static const XMLCh NotBefore[];
            static const XMLCh NotOnOrAfter[];
            static const XMLCh Query[];
            static const XMLCh Recipient[];
            static const XMLCh Request[];
            static const XMLCh Requester[];
            static const XMLCh RequestID[];
            static const XMLCh Resource[];
            static const XMLCh Responder[];
            static const XMLCh RespondWith[];
            static const XMLCh Response[];
            static const XMLCh ResponseID[];
            static const XMLCh Statement[];
            static const XMLCh Status[];
            static const XMLCh StatusMessage[];
            static const XMLCh StatusCode[];
            static const XMLCh StatusDetail[];
            static const XMLCh Subject[];
            static const XMLCh saml_Subject[];
            static const XMLCh SubjectConfirmation[];
            static const XMLCh SubjectConfirmationData[];
            static const XMLCh SubjectLocality[];
            static const XMLCh SubjectQuery[];
            static const XMLCh SubjectStatement[];
            static const XMLCh Success[];
            static const XMLCh Value[];
            static const XMLCh VersionMismatch[];

            // SOAP vocabulary
            static const XMLCh Body[];
            static const XMLCh Envelope[];
            static const XMLCh Fault[];
            static const XMLCh faultcode[];
            static const XMLCh faultstring[];
            static const XMLCh Header[];
            static const XMLCh mustUnderstand[];

            // Signature vocabulary
            static const XMLCh KeyInfo[];
            static const XMLCh Signature[];
            static const XMLCh X509Data[];
            static const XMLCh X509Certificate[];

            // XML vocabulary
            static const XMLCh type[];
            static const XMLCh xml[];
            static const XMLCh xmlns[];
            static const XMLCh xsd[];
            static const XMLCh xsi[];
            static const XMLCh xmlns_xsd[];
            static const XMLCh xmlns_xsi[];
            static const XMLCh xmlns_saml[];
            static const XMLCh xmlns_samlp[];
            static const XMLCh xmlns_soap[];
            static const XMLCh xmlns_code[];
            static const XMLCh xmlns_kind[];
            static const XMLCh xmlns_rw[];
            static const XMLCh xmlns_typens[];
            static const XMLCh typens[];
            static const XMLCh xsi_type[];
            
            // OpenSAML vocabulary
            static const XMLCh ExceptionClass[];
        };

        static void registerSchema(const XMLCh* xmlns, const XMLCh* systemId, DOMEntityResolver* resolver=NULL);

        class SAML_EXPORTS ParserPool : public DOMEntityResolver, DOMErrorHandler
        {
        public:
            ParserPool();
            ~ParserPool();

            DOMBuilder* get();
            void put(DOMBuilder* p);

            void registerSchema(const XMLCh* xmlns, const XMLCh* systemId, DOMEntityResolver* resolver=NULL);

            DOMInputSource* resolveEntity(const XMLCh* const publicId, const XMLCh* const systemId, const XMLCh* const baseURI);
            bool handleError(const DOMError& e);

        private:
#ifdef HAVE_GOOD_STL
            typedef std::map<xstring,xstring> SchemaLocationMap;
            typedef std::map<xstring,DOMEntityResolver*> ResolverMap;
            xstring m_schemaLocations;
#else
            typedef std::map<std::string,std::string> SchemaLocationMap;
            typedef std::map<std::string,DOMEntityResolver*> ResolverMap;
            std::string m_schemaLocations;
#endif
            SchemaLocationMap m_SchemaLocMap;
            ResolverMap m_ResolverMap;
            std::stack<DOMBuilder*> m_pool;
            void* m_lock;
        };

        class SAML_EXPORTS Parser
        {
        public:
            Parser();
            ~Parser();
            DOMDocument* parse(DOMInputSource& domsrc)
            {
                DOMDocument* doc=m_parser->parse(domsrc);
                m_parser->setFeature(XMLUni::fgXercesUserAdoptsDOMDocument,true);
                return doc;
            }

        private:
            Parser(const Parser&) {}
            Parser& operator=(const Parser&) { return *this; }
            DOMBuilder* m_parser;
        };

        class SAML_EXPORTS StreamInputSource : public InputSource
        {
        public:
            StreamInputSource(std::istream& is, const char* systemId=NULL) : InputSource(systemId), m_is(is) {}
            virtual BinInputStream* makeStream() const { return new StreamBinInputStream(m_is); }

        private:
            std::istream& m_is;

            class SAML_EXPORTS StreamBinInputStream : public BinInputStream
            {
            public:
                StreamBinInputStream(std::istream& is) : m_is(is), m_pos(0) {}
                virtual unsigned int curPos() const { return m_pos; }
                virtual unsigned int readBytes(XMLByte* const toFill, const unsigned int maxToRead);
            private:
                std::istream& m_is;
                unsigned int m_pos;
            };
        };
    };

    // Captures global configuration data that must be provided by library client
    class SAML_EXPORTS SAMLConfig
    {
    public:
        SAMLConfig() : compatibility_mode(false), strict_dom_checking(true), clock_skew_secs(300), conn_pool_max(256) {}

        // enables runtime and clients to access global configuration object
        static SAMLConfig& getConfig();

        // global per-process setup and shutdown of runtime
        virtual bool init()=0;
        virtual void term()=0;

        struct SAMLBindingConfig
        {
            SAMLBindingConfig() : timeout(-1), conn_timeout(-1), ssl_ctx_callback(NULL), ssl_ctx_data(NULL) {}
            long timeout;
            long conn_timeout;
            std::string ssl_certfile;
            std::string ssl_certtype;
            std::string ssl_keyfile;
            std::string ssl_keytype;
            std::string ssl_keypass;
            std::string ssl_calist;
            std::string ssl_cadir;
            typedef bool (*ssl_ctx_callback_fn)(void* ssl_ctx, void* userptr);
            ssl_ctx_callback_fn ssl_ctx_callback;
            void* ssl_ctx_data;
        };

    /* start of externally supplied configuration */
        bool compatibility_mode;
        std::string schema_dir;
        std::string log_config;
        std::string inclusive_namespace_prefixes;
        bool strict_dom_checking;
        time_t clock_skew_secs;
        long conn_pool_max;
        SAMLBindingConfig binding_defaults;
    /* end of externally supplied configuration */

        // global mutex available to library applications
        virtual void saml_lock() const=0;
        virtual void saml_unlock() const=0;

        // cross-platform extension library loader
        // extensions must provide:
        //      extern "C" int saml_extension_init(void* context);
        //      extern "C" void saml_extension_term();
        virtual void saml_register_extension(const char* path, void* context=NULL) const=0;
    };


    // Most of the rest of the classes implement SAML constructs. The root
    // object provides a framework for serialization/deserialization and memory
    // management that is tied to the XML parser's behavior and assumptions.
    // Extension classes should always inherit from SAMLObject and use the
    // built-in classes as examples of how to code new types.

    class SAML_EXPORTS SAMLObject
    {
    public:
        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;
        XMLByte* toBase64(unsigned int* outputLength) const;
        virtual ~SAMLObject();
        virtual void checkValidity() const {}
        virtual SAMLObject* clone() const=0;

        DOMDocument* setDocument(DOMDocument* doc) {DOMDocument* ret=m_document; m_document=doc; return ret;}

    protected:
        SAMLObject();
        SAMLObject(std::istream& in);
        SAMLObject(const SAMLObject& src) : m_classname(src.m_classname), m_root(NULL), m_document(NULL), m_bOwnStrings(true), m_log(NULL) {}
        SAMLObject& operator=(const SAMLObject&) { return *this; }

        virtual void fromDOM(DOMElement* e);
        DOMNode* plantRoot();
        void _RTTI(const char* classname);

        friend SAML_EXPORTS std::ostream& operator<<(std::ostream& ostr, const SAMLObject& obj);

        mutable DOMNode* m_root;              // no memory management on DOM itself
        mutable DOMDocument* m_document;      // DOM built manually is freed with "document->release()"
        bool m_bOwnStrings;                   // Does this SAML object own its component fields?
        void* m_log;                          // Logging object, does not expose the implementation
        std::string m_classname;
    };

    // Serialization of objects and wide strings (eventually will use EXC-C14N)
    SAML_EXPORTS std::ostream& operator<<(std::ostream& ostr, const DOMNode& node);
    SAML_EXPORTS std::ostream& operator<<(std::ostream& ostr, const SAMLObject& obj);
    SAML_EXPORTS std::ostream& xmlout(std::ostream& target, const XMLCh* s);

    // Wraps all internal library errors and reflects SAML status structure
    class SAML_EXPORTS SAMLException;
    extern "C" { typedef SAMLException* SAMLExceptionFactory(DOMElement*); }

    class SAML_EXPORTS SAMLException : public std::exception, public SAMLObject
    {
    public:
        static saml::QName SUCCESS;
        static saml::QName REQUESTER;
        static saml::QName RESPONDER;
        static saml::QName VERSIONMISMATCH;

        static SAMLException* getInstance(DOMElement* e);
        static SAMLException* getInstance(std::istream& in);
        static void regFactory(const char* exceptionClass, SAMLExceptionFactory* factory);
        static void unregFactory(const char* exceptionClass);

        SAMLException(const char* msg, DOMElement* detail=NULL);
        SAMLException(const std::string& msg, DOMElement* detail=NULL);
        SAMLException(const Iterator<saml::QName>& codes, const char* msg, DOMElement* detail=NULL);
        SAMLException(const Iterator<saml::QName>& codes, const std::string& msg, DOMElement* detail=NULL);
        SAMLException(const saml::QName& code, const char* msg, DOMElement* detail=NULL);
        SAMLException(const saml::QName& code, const std::string& msg, DOMElement* detail=NULL);
        virtual ~SAMLException() throw ();

        virtual const char* what() const throw () {return (m_msg.c_str());}
        Iterator<saml::QName> getCodes() const {return m_codes;}
        DOMElement* getDetail() const {return m_detail;}

        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;
        virtual SAMLObject* clone() const;

    protected:
        SAMLException(DOMElement* e);
        SAMLException(std::istream& in);
        virtual void fromDOM(DOMElement* e);

    private:
        std::vector<saml::QName> m_codes;
        std::string m_msg;
        mutable DOMElement* m_detail;

        typedef std::map<std::string,SAMLExceptionFactory*> SAMLExceptionFactoryMap;
        static SAMLExceptionFactoryMap m_map;
    };

    // Simplify derivation of exception types
    #define DECLARE_SAML_EXCEPTION(name,base) \
    class SAML_EXPORTS name : public base \
    { \
    public: \
        name(const char* msg) : base(msg) {RTTI(name);} \
        name(const std::string& msg) : base(msg) {RTTI(name);} \
        name(const Iterator<saml::QName>& codes, const char* msg) : base(codes,msg) {RTTI(name);} \
        name(const Iterator<saml::QName>& codes, const std::string& msg) : base(codes, msg) {RTTI(name);} \
        name(const saml::QName& code, const char* msg) : base(code,msg) {RTTI(name);} \
        name(const saml::QName& code, const std::string& msg) : base(code, msg) {RTTI(name);} \
        name(DOMElement* e) : base(e) {RTTI(name);} \
        name(std::istream& in) : base(in) {RTTI(name);} \
        virtual ~name() throw() {} \
    }
    
    #define SAML_EXCEPTION_FACTORY(name) \
        extern "C" SAMLException* name##Factory(DOMElement* e) \
        { \
            return new name(e); \
        }
    
    #define REGISTER_EXCEPTION_FACTORY(name) SAMLException::regFactory("org.opensaml."#name,name##Factory)

    DECLARE_SAML_EXCEPTION(MalformedException,SAMLException);
    DECLARE_SAML_EXCEPTION(UnsupportedExtensionException,SAMLException);
    DECLARE_SAML_EXCEPTION(InvalidCryptoException,SAMLException);
    DECLARE_SAML_EXCEPTION(TrustException,InvalidCryptoException);
    DECLARE_SAML_EXCEPTION(BindingException,SAMLException);
    DECLARE_SAML_EXCEPTION(SOAPException,BindingException);
    DECLARE_SAML_EXCEPTION(ContentTypeException,BindingException);
    DECLARE_SAML_EXCEPTION(UnknownAssertionException,BindingException);

    class SAML_EXPORTS SAMLSignedObject : public SAMLObject
    {
    public:
        virtual ~SAMLSignedObject();

        virtual void sign(signatureMethod alg, XSECCryptoKey* k,
                          const Iterator<XSECCryptoX509*>& certs=EMPTY(XSECCryptoX509*));
        virtual void unsign();
        virtual void verify(XSECCryptoKey* k=NULL) const;
        virtual void verify(XSECCryptoX509& cert) const;
        virtual void verify(XSECKeyInfoResolver* r) const;
        virtual unsigned int getX509CertificateCount() const;
        virtual const XMLCh* getX509Certificate(unsigned int index) const;
        virtual signatureMethod getSignatureAlgorithm() const;
        bool isSigned() const;

    protected:
        SAMLSignedObject(std::istream& in);
        SAMLSignedObject() : m_signature(NULL), m_sigElement(NULL) { RTTI(SAMLSignedObject); }
        
        virtual const XMLCh* getId() const=0;
        virtual void insertSignature()=0;
        DOMElement* getSignatureElement() { return m_sigElement; }

        DSIGSignature* m_signature;
        DOMElement* m_sigElement;
    };

    class SAML_EXPORTS SAMLNameIdentifier : public SAMLObject
    {
    public:
        SAMLNameIdentifier(const XMLCh* name, const XMLCh* nameQualifier=NULL, const XMLCh* format=NULL);
        SAMLNameIdentifier(DOMElement* e);
        SAMLNameIdentifier(std::istream& in);
        virtual ~SAMLNameIdentifier();
        virtual SAMLObject* clone() const;
        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

        const XMLCh* getName() const {return m_name;}
        const XMLCh* getNameQualifier() const {return m_nameQualifier;}
        const XMLCh* getFormat() const {return m_format;}

        static const XMLCh UNSPECIFIED[];
        static const XMLCh EMAIL[];
        static const XMLCh X509[];
        static const XMLCh WINDOWS[];

    protected:
        virtual void fromDOM(DOMElement* e);

        const XMLCh* m_name;
        const XMLCh* m_nameQualifier;
        const XMLCh* m_format;

    private:
        SAMLNameIdentifier(const SAMLNameIdentifier&) {}
        SAMLNameIdentifier& operator=(const SAMLNameIdentifier&) {return *this;}
    };

    class SAML_EXPORTS SAMLSubject : public SAMLObject
    {
    public:
        SAMLSubject(saml::SAMLNameIdentifier* name, const Iterator<const XMLCh*>& confirmationMethods=EMPTY(const XMLCh*),
                    const XMLCh* confirmationData=NULL, DOMElement* keyInfo=NULL);
        SAMLSubject(DOMElement* e);
        SAMLSubject(std::istream& in);
        virtual ~SAMLSubject();
        virtual SAMLObject* clone() const;
        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

        const SAMLNameIdentifier* getNameIdentifier() const {return m_name;}
        Iterator<const XMLCh*> getConfirmationMethods() const {return m_confirmationMethods;}
        const XMLCh* getConfirmationData() const {return m_confirmationData;}
        DOMElement* getKeyInfo() const {return m_keyInfo;}
        
        static const XMLCh ARTIFACT01[];
        static const XMLCh ARTIFACT[];
        static const XMLCh CONF_BEARER[];
        static const XMLCh HOLDER_KEY[];
        static const XMLCh SENDER_VOUCHES[];

    protected:
        virtual void fromDOM(DOMElement* e);

        SAMLNameIdentifier* m_name;
        std::vector<const XMLCh*> m_confirmationMethods;
        const XMLCh* m_confirmationData;
        DOMElement* m_keyInfo;

    private:
        SAMLSubject(const SAMLSubject&) {}
        SAMLSubject& operator=(const SAMLSubject&) {return *this;}
    };

    class SAML_EXPORTS SAMLAttributeDesignator : public SAMLObject
    {
    public:
        SAMLAttributeDesignator() : m_name(NULL), m_namespace(NULL) {RTTI(SAMLAttributeDesignator);}
        SAMLAttributeDesignator(const XMLCh* name, const XMLCh* ns);
        SAMLAttributeDesignator(DOMElement* e);
        SAMLAttributeDesignator(std::istream& in);
        virtual ~SAMLAttributeDesignator();
        virtual void checkValidity() const;
        virtual SAMLObject* clone() const;

        virtual const XMLCh* getName() const {return m_name;}
        virtual const XMLCh* getNamespace() const {return m_namespace;}

        virtual void setName(const XMLCh* name);
        virtual void setNamespace(const XMLCh* ns);

        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

    protected:
        virtual void fromDOM(DOMElement* e);

        const XMLCh* m_name;
        const XMLCh* m_namespace;

    private:
        SAMLAttributeDesignator(const SAMLAttributeDesignator&) {}
        SAMLAttributeDesignator& operator=(const SAMLAttributeDesignator&) {return *this;}
    };

    class SAML_EXPORTS SAMLAttribute;
    extern "C" { typedef SAMLAttribute* SAMLAttributeFactory(DOMElement* e); }

    class SAML_EXPORTS SAMLAttribute : public SAMLObject
    {
    public:
        SAMLAttribute() : m_name(NULL), m_namespace(NULL), m_type(NULL), m_lifetime(0) {RTTI(SAMLAttribute);}
        SAMLAttribute(const XMLCh* name, const XMLCh* ns, const saml::QName* type=NULL, long lifetime=0,
                      const Iterator<const XMLCh*>& values=Iterator<const XMLCh*>());
        SAMLAttribute(DOMElement* e);
        SAMLAttribute(std::istream& in);
        virtual ~SAMLAttribute();
        virtual void checkValidity() const;
        virtual SAMLObject* clone() const;

        virtual const XMLCh* getName() const {return m_name;}
        virtual const XMLCh* getNamespace() const {return m_namespace;}
        virtual const QName* getType() const {return m_type;}
        virtual long getLifetime() const {return m_lifetime;}

        virtual void setName(const XMLCh* name);
        virtual void setNamespace(const XMLCh* ns);
        virtual void setType(const QName* type);
        virtual void setLifetime(long lifetime);

        virtual Iterator<const XMLCh*> getValues() const {return m_values;}
        virtual Iterator<std::string> getSingleByteValues() const;
        virtual Iterator<const DOMElement*> getValueElements() const {return m_elements;}
        virtual void setValues(const Iterator<const XMLCh*>& values=EMPTY(const XMLCh*));
        virtual void addValue(const XMLCh* value);
        virtual void removeValue(unsigned long index);
        
        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

        static SAMLAttributeFactory* setFactory(SAMLAttributeFactory* factory);
        static SAMLAttribute* getInstance(DOMElement* e);
        static SAMLAttribute* getInstance(std::istream& in);

    protected:
        virtual void fromDOM(DOMElement* e);
        virtual void valueToDOM(unsigned int index, DOMElement* e) const;
        virtual XMLCh* computeTypeDecl(DOMElement* e) const;

        const XMLCh* m_name;
        const XMLCh* m_namespace;
        QName* m_type;
        long m_lifetime;
        mutable std::vector<const DOMElement*> m_elements;
        mutable std::vector<const XMLCh*> m_values;
        mutable std::vector<std::string> m_sbValues;

    private:
        SAMLAttribute(const SAMLAttribute&) {}
        SAMLAttribute& operator=(const SAMLAttribute&) {return *this;}

        static SAMLAttributeFactory* m_factory;
    };

    class SAML_EXPORTS SAMLAuthorityBinding : public SAMLObject
    {
    public:
        SAMLAuthorityBinding(const saml::QName& kind, const XMLCh* binding, const XMLCh* location);
        SAMLAuthorityBinding(DOMElement* e);
        SAMLAuthorityBinding(std::istream& in);
        virtual ~SAMLAuthorityBinding();
        virtual SAMLObject* clone() const;

        const saml::QName* getAuthorityKind() const {return m_kind;}
        const XMLCh* getBinding() const {return m_binding;}
        const XMLCh* getLocation() const {return m_location;}

        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

    protected:
        virtual void fromDOM(DOMElement* e);

        saml::QName* m_kind;
        const XMLCh* m_binding;
        const XMLCh* m_location;

    private:
        SAMLAuthorityBinding(const SAMLAuthorityBinding&) {}
        SAMLAuthorityBinding& operator=(const SAMLAuthorityBinding&) {return *this;}
    };

    class SAML_EXPORTS SAMLStatement;
    extern "C" { typedef SAMLStatement* SAMLStatementFactory(DOMElement*); }

    class SAML_EXPORTS SAMLStatement : public SAMLObject
    {
    public:
        virtual ~SAMLStatement() {}

        static void regFactory(saml::QName& type, SAMLStatementFactory* factory) {m_map.insert(SAMLStatementFactoryMap::value_type(type,factory));}
        static void unregFactory(saml::QName& type) {m_map.erase(type);}
        static SAMLStatement* getInstance(DOMElement* e);
        static SAMLStatement* getInstance(std::istream& in);

    protected:
        SAMLStatement() { RTTI(SAMLStatement); }
        SAMLStatement(std::istream& in) : SAMLObject(in) { RTTI(SAMLStatement); }

    private:
        SAMLStatement(const SAMLStatement&) {}
        SAMLStatement& operator=(const SAMLStatement&) {return *this;}

        typedef std::map<saml::QName,SAMLStatementFactory*> SAMLStatementFactoryMap;
        static SAMLStatementFactoryMap m_map;
    };

    class SAML_EXPORTS SAMLCondition;
    extern "C" { typedef SAMLCondition* SAMLConditionFactory(DOMElement*); }

    class SAML_EXPORTS SAMLCondition : public SAMLObject
    {
    public:
        virtual ~SAMLCondition() {}

        static void regFactory(saml::QName& type, SAMLConditionFactory* factory) {m_map.insert(SAMLConditionFactoryMap::value_type(type,factory));}
        static void unregFactory(saml::QName& type) {m_map.erase(type);}
        static SAMLCondition* getInstance(DOMElement* e);
        static SAMLCondition* getInstance(std::istream& in);

    protected:
        SAMLCondition() { RTTI(SAMLCondition); }
        SAMLCondition(std::istream& in) : SAMLObject(in) { RTTI(SAMLCondition); }

    private:
        SAMLCondition(const SAMLCondition&) {}
        SAMLCondition& operator=(const SAMLCondition&) {return *this;}

        typedef std::map<saml::QName,SAMLConditionFactory*> SAMLConditionFactoryMap;
        static SAMLConditionFactoryMap m_map;
    };

    class SAML_EXPORTS SAMLAudienceRestrictionCondition : public SAMLCondition
    {
    public:
        SAMLAudienceRestrictionCondition(const Iterator<const XMLCh*>& audiences);
        SAMLAudienceRestrictionCondition(DOMElement* e);
        SAMLAudienceRestrictionCondition(std::istream& in);
        virtual ~SAMLAudienceRestrictionCondition();
        virtual SAMLObject* clone() const;

        saml::Iterator<const XMLCh*> getAudiences() {return m_audiences;}

        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

        bool eval(const Iterator<const XMLCh*>& audiences) const;
#ifdef HAVE_GOOD_STL
        bool eval(const Iterator<xstring>& audiences) const;
#endif

    protected:
        virtual void fromDOM(DOMElement* e);

        std::vector<const XMLCh*> m_audiences;

    private:
        SAMLAudienceRestrictionCondition(const SAMLAudienceRestrictionCondition&) {}
        SAMLAudienceRestrictionCondition& operator=(const SAMLAudienceRestrictionCondition&) {return *this;}
    };

    class SAML_EXPORTS SAMLDoNotCacheCondition : public SAMLCondition
    {
    public:
        SAMLDoNotCacheCondition();
        SAMLDoNotCacheCondition(DOMElement* e);
        SAMLDoNotCacheCondition(std::istream& in);
        virtual ~SAMLDoNotCacheCondition() {}
        virtual SAMLObject* clone() const;

        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

    protected:
        virtual void fromDOM(DOMElement* e);

    private:
        SAMLDoNotCacheCondition(const SAMLDoNotCacheCondition&) {}
        SAMLDoNotCacheCondition& operator=(const SAMLDoNotCacheCondition&) {return *this;}
    };

    class SAML_EXPORTS SAMLAssertion : public SAMLSignedObject
    {
    public:
        SAMLAssertion(const XMLCh* issuer, const XMLDateTime* notBefore=NULL, const XMLDateTime* notOnOrAfter=NULL,
                      const Iterator<SAMLCondition*>& conditions=Iterator<SAMLCondition*>(),
                      const Iterator<SAMLStatement*>& statements=Iterator<SAMLStatement*>());
        SAMLAssertion(DOMElement* e);
        SAMLAssertion(std::istream& in);
        virtual ~SAMLAssertion();
        virtual void checkValidity() const;
        virtual SAMLObject* clone() const;

        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

        const XMLCh* getId() const {return m_assertionId;}
        const XMLCh* getIssuer() const {return m_issuer;}
        const XMLDateTime* getIssueInstant() const {return m_issueInstant;}
        const XMLDateTime* getNotBefore() const {return m_notBefore;}
        const XMLDateTime* getNotOnOrAfter() const {return m_notOnOrAfter;}
        Iterator<SAMLCondition*> getConditions() const {return m_conditions;}
        Iterator<SAMLStatement*> getStatements() const {return m_statements;}

        void setStatements(const Iterator<SAMLStatement*>& statements=EMPTY(SAMLStatement*));
        void addStatement(SAMLStatement* statement);
        void removeStatement(unsigned long index);

    protected:
        virtual void fromDOM(DOMElement* e);
        virtual void insertSignature();

        mutable const XMLCh* m_assertionId;
        const XMLCh* m_issuer;
        XMLDateTime* m_issueInstant;
        XMLDateTime* m_notBefore;
        XMLDateTime* m_notOnOrAfter;
        std::vector<SAMLCondition*> m_conditions;
        std::vector<SAMLStatement*> m_statements;

    private:
        SAMLAssertion(const SAMLAssertion&) {}
        SAMLAssertion& operator=(const SAMLAssertion&) {return *this;}
    };

    class SAML_EXPORTS SAMLSubjectStatement : public SAMLStatement
    {
    public:
        virtual ~SAMLSubjectStatement();
        virtual void checkValidity() const;
        virtual SAMLObject* clone() const;

        const SAMLSubject* getSubject() const {return m_subject;}
        void setSubject(SAMLSubject* subject);

    protected:
        SAMLSubjectStatement() {RTTI(SAMLSubjectStatement);}
        SAMLSubjectStatement(SAMLSubject* subject);
        SAMLSubjectStatement(DOMElement* e);
        SAMLSubjectStatement(std::istream& in);
        virtual void fromDOM(DOMElement* e);
        SAMLSubject* m_subject;

    private:
        SAMLSubjectStatement(const SAMLSubjectStatement&) {}
        SAMLSubjectStatement& operator=(const SAMLSubjectStatement&) {return *this;}
    };

    class SAML_EXPORTS SAMLAttributeStatement : public SAMLSubjectStatement
    {
    public:
        SAMLAttributeStatement() {RTTI(SAMLAttributeStatement);}
        SAMLAttributeStatement(SAMLSubject* subject, const Iterator<SAMLAttribute*>& attributes=Iterator<SAMLAttribute*>());
        SAMLAttributeStatement(DOMElement* e);
        SAMLAttributeStatement(std::istream& in);
        virtual ~SAMLAttributeStatement();
        virtual void checkValidity() const;
        virtual SAMLObject* clone() const;

        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

        Iterator<SAMLAttribute*> getAttributes() const {return m_attributes;}
        void setAttributes(const Iterator<SAMLAttribute*>& attributes=EMPTY(SAMLAttribute*));
        void addAttribute(SAMLAttribute* attribute);
        void removeAttribute(unsigned long index);

    protected:
        virtual void fromDOM(DOMElement* e);
        std::vector<SAMLAttribute*> m_attributes;

    private:
        SAMLAttributeStatement(const SAMLAttributeStatement&) {}
        SAMLAttributeStatement& operator=(const SAMLAttributeStatement&) {return *this;}
    };

    class SAML_EXPORTS SAMLAuthenticationStatement : public SAMLSubjectStatement
    {
    public:
        SAMLAuthenticationStatement(SAMLSubject* subject,
                                    const XMLCh* authMethod, const XMLDateTime& authInstant,
                                    const XMLCh* subjectIP=NULL, const XMLCh* subjectDNS=NULL,
                                    const Iterator<SAMLAuthorityBinding*>& bindings=Iterator<SAMLAuthorityBinding*>());
        SAMLAuthenticationStatement(DOMElement* e);
        SAMLAuthenticationStatement(std::istream& in);
        virtual ~SAMLAuthenticationStatement();
        virtual SAMLObject* clone() const;

        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

        const XMLCh* getSubjectIP() const {return m_subjectIP;}
        const XMLCh* getSubjectDNS() const {return m_subjectDNS;}
        const XMLCh* getAuthMethod() const {return m_authMethod;}
        const XMLDateTime* getAuthInstant() const {return m_authInstant;}
        Iterator<SAMLAuthorityBinding*> getBindings() const {return m_bindings;}

    protected:
        virtual void fromDOM(DOMElement* e);

        const XMLCh* m_subjectIP;
        const XMLCh* m_subjectDNS;
        const XMLCh* m_authMethod;
        XMLDateTime* m_authInstant;
        std::vector<SAMLAuthorityBinding*> m_bindings;

    private:
        SAMLAuthenticationStatement(const SAMLAuthenticationStatement&) {}
        SAMLAuthenticationStatement& operator=(const SAMLAuthenticationStatement&) {return *this;}
    };

    class SAML_EXPORTS SAMLAction : public SAMLObject
    {
    public:
        SAMLAction(const XMLCh* data, const XMLCh* nsURI = NULL);
        SAMLAction(DOMElement* e);
        SAMLAction(std::istream& in);
        virtual ~SAMLAction();
    
        virtual void checkValidity() const;
        virtual SAMLObject* clone() const;
        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

        static const XMLCh SAML_ACTION_NAMESPACE_RWEDC_NEG[];
        static const XMLCh SAML_ACTION_NAMESPACE_RWEDC[];
        static const XMLCh SAML_ACTION_NAMESPACE_GHPP[];
        static const XMLCh SAML_ACTION_NAMESPACE_UNIX[];
    
        const XMLCh* getNamespace() const {return m_namespace ? m_namespace : SAML_ACTION_NAMESPACE_RWEDC_NEG;}
        const XMLCh* getData() const {return m_data;}
    
    protected:
        virtual void fromDOM(DOMElement* e);
    
        const XMLCh* m_namespace;
        const XMLCh* m_data;

    private:
        SAMLAction(const SAMLAction&) {}
        SAMLAction& operator=(const SAMLAction&) {return *this;}
    };

    class SAML_EXPORTS SAMLEvidence : public SAMLObject
    {
    public:
        SAMLEvidence(const Iterator<SAMLAssertion*>& assertions, const Iterator<const XMLCh*>& assertionIDRefs=EMPTY(const XMLCh*));
        SAMLEvidence(const Iterator<const XMLCh*>& assertionIDRefs, const Iterator<SAMLAssertion*>& assertions=EMPTY(SAMLAssertion*));
        SAMLEvidence(DOMElement* e);
        SAMLEvidence(std::istream& in);
        virtual ~SAMLEvidence();
    
        virtual void checkValidity() const;
        virtual SAMLObject* clone() const;
        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;
    
        Iterator<SAMLAssertion*> getAssertions() const {return m_assertions;}
        Iterator<const XMLCh*> getAssertionIDRefs() const {return m_assertionIDRefs;}

    protected:
        virtual void fromDOM(DOMElement* e);
    
        std::vector<SAMLAssertion*> m_assertions;
        std::vector<const XMLCh*> m_assertionIDRefs;

    private:
        void construct(const Iterator<SAMLAssertion*>& assertions, const Iterator<const XMLCh*>& assertionIDRefs);

        SAMLEvidence(const SAMLEvidence&) {}
        SAMLEvidence& operator=(const SAMLEvidence&) {return *this;}
    };

    class SAML_EXPORTS SAMLDecision
    {
    public:
        static const XMLCh Permit[];
        static const XMLCh Deny[];
        static const XMLCh Indeterminate[];
    };

    class SAML_EXPORTS SAMLAuthorizationDecisionStatement : public SAMLSubjectStatement
    {
    public:
        SAMLAuthorizationDecisionStatement(
            SAMLSubject* subject, const XMLCh* resource, const XMLCh* decision, const Iterator<SAMLAction*>& actions, SAMLEvidence* evidence=NULL
            );
        SAMLAuthorizationDecisionStatement(DOMElement* e);
        SAMLAuthorizationDecisionStatement(std::istream& in);
        virtual ~SAMLAuthorizationDecisionStatement();
    
        virtual void checkValidity() const;
        virtual SAMLObject* clone() const;
        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;
    
        const XMLCh* getResource() const {return m_resource;}
        const XMLCh* getDecision() const {return m_decision;}
        Iterator<SAMLAction*> getActions() const {return m_actions;}
        SAMLEvidence* getEvidence() const {return m_evidence;}

    protected:
        virtual void fromDOM(DOMElement* e);
    
        const XMLCh* m_resource;
        const XMLCh* m_decision;
        std::vector<SAMLAction*> m_actions;
        SAMLEvidence* m_evidence;

    private:
        SAMLAuthorizationDecisionStatement(const SAMLAuthorizationDecisionStatement&) {}
        SAMLAuthorizationDecisionStatement& operator=(const SAMLAuthorizationDecisionStatement&) {return *this;}
    };

    class SAML_EXPORTS SAMLQuery;
    extern "C" { typedef SAMLQuery* SAMLQueryFactory(DOMElement*); }

    class SAML_EXPORTS SAMLQuery : public SAMLObject
    {
    public:
        virtual ~SAMLQuery() {}

        static void regFactory(saml::QName& type, SAMLQueryFactory* factory) {m_map.insert(SAMLQueryFactoryMap::value_type(type,factory));}
        static void unregFactory(saml::QName& type) {m_map.erase(type);}
        static SAMLQuery* getInstance(DOMElement* e);
        static SAMLQuery* getInstance(std::istream& in);

    protected:
        SAMLQuery() { RTTI(SAMLQuery); }
        SAMLQuery(std::istream& in) : SAMLObject(in) { RTTI(SAMLQuery); }

    private:
        SAMLQuery(const SAMLQuery&) {}
        SAMLQuery& operator=(const SAMLQuery&) {return *this;}

        typedef std::map<saml::QName,SAMLQueryFactory*> SAMLQueryFactoryMap;
        static SAMLQueryFactoryMap m_map;
    };

    class SAML_EXPORTS SAMLSubjectQuery : public SAMLQuery
    {
    public:
        virtual ~SAMLSubjectQuery();
        virtual void checkValidity() const;
        virtual SAMLObject* clone() const;

        const SAMLSubject* getSubject() const {return m_subject;}
        void setSubject(SAMLSubject* subject);

    protected:
        SAMLSubjectQuery() {RTTI(SAMLSubjectQuery);}
        SAMLSubjectQuery(SAMLSubject* subject);
        SAMLSubjectQuery(DOMElement* e);
        SAMLSubjectQuery(std::istream& in);
        virtual void fromDOM(DOMElement* e);
        SAMLSubject* m_subject;

    private:
        SAMLSubjectQuery(const SAMLSubjectQuery&) {}
        SAMLSubjectQuery& operator=(const SAMLSubjectQuery&) {return *this;}
    };

    class SAML_EXPORTS SAMLAttributeQuery : public SAMLSubjectQuery
    {
    public:
        SAMLAttributeQuery(SAMLSubject* subject, const XMLCh* resource=NULL,
                           const Iterator<SAMLAttributeDesignator*>& designators=Iterator<SAMLAttributeDesignator*>());
        SAMLAttributeQuery(DOMElement* e);
        SAMLAttributeQuery(std::istream& in);
        virtual ~SAMLAttributeQuery();
        virtual SAMLObject* clone() const;

        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

        const XMLCh* getResource() const {return m_resource;}
        Iterator<SAMLAttributeDesignator*> getDesignators() const {return m_designators;}

    protected:
        virtual void fromDOM(DOMElement* e);

        const XMLCh* m_resource;
        std::vector<SAMLAttributeDesignator*> m_designators;

    private:
        SAMLAttributeQuery(const SAMLAttributeQuery&) {}
        SAMLAttributeQuery& operator=(const SAMLAttributeQuery&) {return *this;}
    };

    class SAML_EXPORTS SAMLAuthenticationQuery : public SAMLSubjectQuery
    {
    public:
        SAMLAuthenticationQuery(SAMLSubject* subject, const XMLCh* authMethod=NULL);
        SAMLAuthenticationQuery(DOMElement* e);
        SAMLAuthenticationQuery(std::istream& in);
        virtual ~SAMLAuthenticationQuery();
    
        virtual SAMLObject* clone() const;
        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;
    
        const XMLCh* getAuthMethod() const {return m_authMethod;}

    protected:
        virtual void fromDOM(DOMElement* e);
    
        const XMLCh* m_authMethod;

    private:
        SAMLAuthenticationQuery(const SAMLAuthenticationQuery&) {}
        SAMLAuthenticationQuery& operator=(const SAMLAuthenticationQuery&) {return *this;}
    };

    class SAML_EXPORTS SAMLAuthorizationDecisionQuery : public SAMLSubjectQuery
    {
    public:
        SAMLAuthorizationDecisionQuery(
            SAMLSubject* subject, const XMLCh* resource, const Iterator<SAMLAction*>& actions, SAMLEvidence* evidence=NULL
            );
        SAMLAuthorizationDecisionQuery(DOMElement* e);
        SAMLAuthorizationDecisionQuery(std::istream& in);
        virtual ~SAMLAuthorizationDecisionQuery();
    
        virtual void checkValidity() const;
        virtual SAMLObject* clone() const;
        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;
    
        const XMLCh* getResource() const {return m_resource;}
        Iterator<SAMLAction*> getActions() const {return m_actions;}
        SAMLEvidence* getEvidence() const {return m_evidence;}

    protected:
        virtual void fromDOM(DOMElement* e);
    
        const XMLCh* m_resource;
        std::vector<SAMLAction*> m_actions;
        SAMLEvidence* m_evidence;

    private:
        SAMLAuthorizationDecisionQuery(const SAMLAuthorizationDecisionQuery&) {}
        SAMLAuthorizationDecisionQuery& operator=(const SAMLAuthorizationDecisionQuery&) {return *this;}
    };

    class SAML_EXPORTS SAMLRequest : public SAMLSignedObject
    {
    public:
        SAMLRequest(const Iterator<saml::QName>& respondWiths=Iterator<saml::QName>(), SAMLQuery* query=NULL,
                    const Iterator<const XMLCh*>& assertionIdRefs=Iterator<const XMLCh*>(),
                    const Iterator<const XMLCh*>& artifacts=Iterator<const XMLCh*>());
        SAMLRequest(DOMElement* e);
        SAMLRequest(std::istream& in);
        virtual ~SAMLRequest();
        virtual SAMLObject* clone() const;

        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

        const XMLCh* getId() const {return m_requestId;}
        const XMLDateTime* getIssueInstant() const {return m_issueInstant;}
        Iterator<saml::QName> getRespondWiths() const {return m_respondWiths;}
        const SAMLQuery* getQuery() const {return m_query;}
        Iterator<const XMLCh*> getAssertionIdRefs() const {return m_assertionIdRefs;}
        Iterator<const XMLCh*> getArtifacts() const {return m_artifacts;}

    protected:
        virtual void fromDOM(DOMElement* e);
        virtual void insertSignature();

        mutable const XMLCh* m_requestId;
        XMLDateTime* m_issueInstant;
        std::vector<saml::QName> m_respondWiths;
        SAMLQuery* m_query;
        std::vector<const XMLCh*> m_assertionIdRefs;
        std::vector<const XMLCh*> m_artifacts;

    private:
        SAMLRequest(const SAMLRequest&) {}
        SAMLRequest& operator=(const SAMLRequest&) {return *this;}
    };

    class SAML_EXPORTS SAMLResponse : public SAMLSignedObject
    {
    public:
        SAMLResponse(const XMLCh* inResponseTo=NULL, const XMLCh* recipient=NULL,
                     const Iterator<SAMLAssertion*>& assertions=Iterator<SAMLAssertion*>(), SAMLException* e=NULL);
        SAMLResponse(DOMElement* e);
        SAMLResponse(std::istream& in);
        virtual ~SAMLResponse();
        virtual SAMLObject* clone() const;

        virtual DOMNode* toDOM(DOMDocument* doc=NULL, bool xmlns=true) const;

        const XMLCh* getId() const {return m_responseId;}
        const XMLCh* getInResponseTo() const {return m_inResponseTo;}
        const XMLDateTime* getIssueInstant() const {return m_issueInstant;}
        const XMLCh* getRecipient() const {return m_recipient;}
        Iterator<SAMLAssertion*> getAssertions() const {return m_assertions;}

        void setAssertions(const Iterator<SAMLAssertion*>& assertions=EMPTY(SAMLAssertion*));
        void addAssertion(SAMLAssertion* assertion);
        void removeAssertion(unsigned long index);

    protected:
        virtual void fromDOM(DOMElement* e);
        virtual void insertSignature();

        mutable const XMLCh* m_responseId;
        const XMLCh* m_inResponseTo;
        XMLDateTime* m_issueInstant;
        const XMLCh* m_recipient;
        std::vector<SAMLAssertion*> m_assertions;
        SAMLException* m_exception;

    private:
        SAMLResponse(const SAMLResponse&) {}
        SAMLResponse& operator=(const SAMLResponse&) {return *this;}
    };


    // The final classes support SAML binding and profile implementations that
    // use the SAML objects to do useful things.

    DECLARE_SAML_EXCEPTION(ProfileException,SAMLException);
    DECLARE_SAML_EXCEPTION(FatalProfileException,ProfileException);
    DECLARE_SAML_EXCEPTION(RetryableProfileException,ProfileException);
    DECLARE_SAML_EXCEPTION(ExpiredAssertionException,FatalProfileException);
    DECLARE_SAML_EXCEPTION(InvalidAssertionException,FatalProfileException);
    DECLARE_SAML_EXCEPTION(ReplayedAssertionException,RetryableProfileException);

    class SAML_EXPORTS SAMLPOSTProfile
    {
    public:
        static SAMLResponse* prepare(
            const XMLCh* recipient,
            const XMLCh* issuer,
            const Iterator<const XMLCh*>& audiences,
            const XMLCh* name,
            const XMLCh* nameQualifier,
            const XMLCh* format,
            const XMLCh* subjectIP,
            const XMLCh* authMethod,
            const XMLDateTime& authInstant,
            const Iterator<SAMLAuthorityBinding*>& bindings
            );
        static SAMLResponse* accept(const XMLByte* buf, const XMLCh* receiver, int ttlSeconds, bool process=true);
        static void process(SAMLResponse& r, const XMLCh* receiver, int ttlSeconds);
        static const SAMLAssertion* getSSOAssertion(const SAMLResponse& r, const Iterator<const XMLCh*>& audiences);
        static const SAMLAuthenticationStatement* getSSOStatement(const SAMLAssertion& a);
        static bool checkReplayCache(const SAMLAssertion& a);

    private:
        static std::set<std::string> m_replayCache;
        static std::multimap<std::string,std::string> m_replayExpMap;
    };

    class SAML_EXPORTS SAMLBinding
    {
    public:
        virtual ~SAMLBinding() {}

        virtual SAMLResponse* send(
            const SAMLAuthorityBinding& bindingInfo, SAMLRequest& req,
            SAMLConfig::SAMLBindingConfig& conf=SAMLConfig::getConfig().binding_defaults
            )=0;
        virtual SAMLRequest* receive(void* reqContext, const XMLCh** requester)=0;
        virtual void respond(void* respContext, SAMLResponse* response, SAMLException* e)=0;

        static const XMLCh SAML_SOAP_HTTPS[];
        static SAMLBinding* getInstance(const XMLCh* binding);
    };

    class SAML_EXPORTS SAMLSOAPBinding : public SAMLBinding
    {
    public:
        SAMLSOAPBinding();
        virtual ~SAMLSOAPBinding();
    
        virtual SAMLResponse* send(
            const SAMLAuthorityBinding& bindingInfo, SAMLRequest& req,
            SAMLConfig::SAMLBindingConfig& conf=SAMLConfig::getConfig().binding_defaults
            );
        virtual SAMLRequest* receive(void* reqContext, const XMLCh** requester);
        virtual void respond(void* respContext, SAMLResponse* response, SAMLException* e);

        static std::string version;     // custom HTTP header
    };
}

#endif
